"""
Contains the class `ModelStructure`.
"""

from typing import Literal, Any

import numpy as np

from .voltagefunction import load_voltage_model
from .basisfunctions import extract_basis_functions, combine_symbols, \
    generate_signals, generate_basis_functions, generate_signal_trajectories
from .regression import setup_regression, run_optimizer
from .dataclasses import VoltageFunction, BasisFunction, Model, Signal

class ModelStructure:
    """This class allows battery model identification and simulation employing
    the proposed model structure in the linear parameter-varying (LPV) framework
    using input-output (IO) representation.
    
    Attributes
    ----------
    battery_capacity : float
        Capacity of the battery being modelled.
    sampling_period: float
        Model sampling time.
    """

    battery_capacity: float
    sampling_period: float

    _emf_function: VoltageFunction
    _hysteresis_function: VoltageFunction | None

    _basis_functions: list[BasisFunction]
    _hysteresis_basis_functions: list[BasisFunction]

    def __init__(self, battery_capacity: float, sampling_period: float):
        self.battery_capacity = battery_capacity
        self.sampling_period = sampling_period

        self._hysteresis_function = None

        self._basis_functions = []
        self._hysteresis_basis_functions = []

        self.models = []

    def add_emf_function(self, soc_values: list[float], voltage_values: list[float]):
        """Add EMF function used to decompose battery voltage into
        overpotentials and vice versa."""
        self._emf_function = load_voltage_model(soc_values, voltage_values)

    def add_hysteresis_function(self, soc_values: list[float], voltage_values: list[float]):
        """Add hysteresis function to be used as second model input."""
        self._hysteresis_function = load_voltage_model(soc_values, voltage_values)

    def add_basis_functions(self, basis_function_strings: list[str],
                            hysteresis_basis_function_strings: list[str] | None = None):
        """Add basis functions by extracting signal variable, the desired
        operation and its arguments."""
        self._basis_functions = extract_basis_functions(basis_function_strings)
        if hysteresis_basis_function_strings is not None:
            hbfs = hysteresis_basis_function_strings
            self._hysteresis_basis_functions = extract_basis_functions(hbfs)

    def simulate(self, model: Model, dataset: dict, initial_soc: float,
                 no_of_initial_values: int | None = None):
        """Simulate the battery voltage using the provided model
        with the battery current as input."""
        #
        if no_of_initial_values is None:
            no_of_initial_values = model.model_order
        #
        signals = generate_signals(dataset, self._emf_function, self._hysteresis_function,
                                   initial_soc, self.sampling_period, self.battery_capacity)
        #
        basis_functions = generate_basis_functions(self._basis_functions, signals)
        hysteresis_basis_functions = generate_basis_functions(self._hysteresis_basis_functions,
                                                              signals)
        #
        io_signals = [ signals.find('v'), signals.find('i') ]
        if self._hysteresis_function is not None:
            io_signals.append( signals.find('h') )
        #
        signal_tuple = io_signals, basis_functions.signals, hysteresis_basis_functions.signals
        signal_trajectories = generate_signal_trajectories(signal_tuple, model.model_order,
                                                           no_of_initial_values)
        io_trajectories, p_trajectories, h_trajectories = signal_trajectories
        #
        simulated_voltage = Signal('v_sim', signals.find('v').trajectory[:no_of_initial_values],
                                   lambda x: x)
        combined_trajectories_dict = io_trajectories | p_trajectories | h_trajectories
        #
        for delay in range(1, model.model_order + 1):
            key = simulated_voltage.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
            #
            combined_trajectories_dict[key] = [simulated_voltage.trajectory[-delay]]

        for k in range( len(signals.find('i').trajectory) - no_of_initial_values ):
            #
            phi = []
            for key in model.model_terms:
                # Replace v with v_sim in the symbols
                new_symbols = []
                for sym in key.split('*'):
                    sym_parts = sym.split('(')
                    if sym_parts[0] == 'v':
                        new_symbols.append(simulated_voltage.symbol + '(' + sym_parts[1])
                    else:
                        new_symbols.append(sym)
                # The value corresponding to each column for kth time instant.
                term_value = np.prod([ combined_trajectories_dict[sym][k] for sym in new_symbols ])
                phi.append(term_value)

            # Newly calculated voltage value at kth time instant.
            v_sim_new_value = np.dot( phi, model.model_estimate )
            # Make updates.
            simulated_voltage.update_trajectory(v_sim_new_value)
            for delay in range(1, model.model_order + 1):
                key = simulated_voltage.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
                #
                combined_trajectories_dict[key].append( simulated_voltage.trajectory[-delay] )

        emf_trajectory = self._emf_function(signals.find('s').trajectory)
        return np.add(emf_trajectory, simulated_voltage.trajectory)

    def identify(self, dataset: dict[Literal['current', 'voltage'], Any], initial_soc: float,
                 model_order: int, nonlinearity_order: int, optimizers: Literal['lasso', 'ridge']):
        """Identify a battery model using the provided identification
        dataset along with the desired model order and basis-function
        complexity."""
        #
        signals = generate_signals(dataset, self._emf_function, self._hysteresis_function,
                                   initial_soc, self.sampling_period, self.battery_capacity)
        #
        basis_functions = generate_basis_functions(self._basis_functions, signals)
        hysteresis_basis_functions = generate_basis_functions(self._hysteresis_basis_functions,
                                                              signals)
        #
        io_signals = [ signals.find('v'), signals.find('i') ]
        if self._hysteresis_function is not None:
            io_signals.append( signals.find('h') )
        #
        signal_tuple = io_signals, basis_functions.signals, hysteresis_basis_functions.signals
        signal_trajectories = generate_signal_trajectories(signal_tuple, model_order, model_order)
        io_trajectories, p_trajectories, h_trajectories = signal_trajectories
        #
        basis_function_keys = combine_symbols(basis_functions.get_symbols(), nonlinearity_order)
        hysteresis_basis_function_keys = hysteresis_basis_functions.get_symbols()
        #
        regression_matrix, regressor_labels = setup_regression(io_trajectories,
                                                               p_trajectories,
                                                               h_trajectories,
                                                               basis_function_keys,
                                                               hysteresis_basis_function_keys)
        output_vector = np.array(signals.find('v').trajectory[model_order:])
        #
        if not optimizers:
            raise ValueError('Unspecified optimization routine(s).')
        #
        model_estimate = np.array([])
        for optimizer in optimizers:
            #
            model_estimate = run_optimizer(regression_matrix, output_vector, optimizer)
            #
            if optimizer == 'lasso':
                selected_regressors = np.abs(model_estimate) > 1e-5
                regression_matrix = regression_matrix[:, selected_regressors] # type: ignore
                regressor_labels = np.array(regressor_labels)[selected_regressors]

        return Model(model_order, nonlinearity_order, regressor_labels, model_estimate)



"""abc"""

import re
import math as m
from enum import Enum
from typing import Callable, Tuple, Any
from itertools import combinations_with_replacement

import numpy as np
from .dataclasses import BasisFunction, Signal, SignalVector, VoltageFunction

class Operation(Enum):
    """List of allowed operations."""
    NONE = 0
    INVERSE = 1
    LOGARITHM = 2
    EXP_SQRT = 3
    LOWPASS = 4

def extract_basis_functions(basis_function_strings: list[str]) -> list[BasisFunction]:
    """Extract basis functions from a list of strings by recognizing
    symbols of various signals and respective operations."""
    basis_functions = []
    # We define tuples for various operations in the format
    # (regex, indices for variable and arguments, operation).
    identifiers = [
        (r"^(s|i|d|v)$", [0], Operation.NONE),
        (r"^(s|i|d|v)⁻¹$", [0], Operation.INVERSE),
        (r"^log\[(s|d|i|v)\]$", [0], Operation.LOGARITHM),
        (r"^exp\[(([0-9]*[.])?[0-9]+)sqrt\[\|(s|d|i|v)\|\]\]$", [2, 0], Operation.EXP_SQRT),
        (r"^(s|i|d|v)\[(([0-9]*[.])?[0-9]+),(([0-9]*[.])?[0-9]+)\]$", [0, 1, 3], Operation.LOWPASS)
    ]
    #
    for function_string in basis_function_strings:
        for identifier, indices, operation in identifiers:
            #
            result = re.findall(identifier, function_string.strip())
            if len(result) > 0:
                result = result[0] if isinstance(result[0], tuple) else result
                args = [ result[i] for i in indices[1:] ]
                basis_functions.append( BasisFunction(result[indices[0]], operation, args) )
                break
    return basis_functions

def create_functional_variant(signal: Signal, sym_func: str, equation: Callable):
    """Create functional variants of the given signals."""
    new_symbol = sym_func
    new_trajectory = [equation(x) for x in signal.trajectory]

    return Signal(new_symbol, new_trajectory, lambda x: x)

def perform_signal_operation(signal: Signal, operation: Operation, args: list[Any]) -> Signal:
    """Perform an operation on a signal."""
    result_signal = signal
    if operation == Operation.INVERSE:
        result_signal = create_functional_variant(signal, f'{signal.symbol}⁻¹', lambda x: 1 / x)
    elif operation == Operation.LOGARITHM:
        result_signal = create_functional_variant(signal, f'log[{signal.symbol}]', m.log)
    elif operation == Operation.EXP_SQRT:
        gamma = float( args[0] )
        result_signal = create_functional_variant(signal, f'exp[{args[0]}sqrt[|{signal.symbol}|]]',
                                                  lambda x: m.exp(gamma * m.sqrt(abs(x) )))
    elif operation == Operation.LOWPASS:
        epsilon = lambda d: float(args[1]) if d == 0 else float(args[0]) # pylint: disable=C3001
        #
        result_signal = Signal(f'{signal.symbol}[{args[0]},{args[1]}]', [signal.trajectory[0]],
                               lambda past_value, d: epsilon(d) * past_value + (1 - epsilon(d)) * d,
                               equation_order=1)
        # Update the trajectory
        for value in signal.trajectory[1:]:
            result_signal.update_trajectory(value)

    return result_signal

def combine_symbols(symbols: list[str], nl_order: int):
    """Use symbols to generate higher-order combinations."""
    #
    combinations = []
    for number_of_terms in range(1, nl_order + 1):
        combinations += list(combinations_with_replacement(symbols, number_of_terms))

    # We develop the conditions to filter unwanted combinations.
    # 1. If a key has more than one delta symbols
    def more_than_one_delta(key):
        return [ k[0] == 'd' for k in key ].count(True) > 1
    # 2. If a key has `s` and its inverse.
    def s_and_inverse_s(key):
        return 's' in key and 's⁻¹' in key
    # We combine the above two conditions
    def check_allowed(key):
        return not more_than_one_delta(key) and not s_and_inverse_s(key)

    # We filter out unnecessary combinations.
    final_combinations = filter( check_allowed, combinations )
    return list(final_combinations)

def generate_signals(dataset: dict, emf_function: VoltageFunction,
                     hysteresis_function: VoltageFunction | None, initial_soc: float,
                     sampling_period: float, battery_capacity: float):
    """Generate basic signals, that is, SOC, current, voltage,
    current direction, and hysteresis function using
    the `Signal` class."""
    # First, we generate input--output signal trajectories
    # 1. SOC trajectory
    soc = Signal('s', [initial_soc],
                    lambda soc, current, delta_t, capacity: soc + delta_t / capacity * current,
                    equation_order=1)
    # Update the trajectory
    for current in dataset['current'][:-1]:
        soc.update_trajectory(current, sampling_period, battery_capacity)
    # 2. Current trajectory
    current = Signal('i', dataset['current'], lambda x: x)
    # 3. Overpotential trajectory
    emf_trajectory = emf_function(soc.trajectory)
    overpotential = Signal('v', (dataset['voltage'] - emf_trajectory).tolist(), lambda x: x)
    # 4. Basic current-direction trajectory
    current_direction =  Signal('d', np.sign(current.trajectory).tolist(), lambda x: x)
    #
    signals = SignalVector([ soc, current, overpotential, current_direction ])
    # 5. Hysteresis-input
    if hysteresis_function is not None:
        hysteresis_input = Signal('h', hysteresis_function(soc.trajectory),
                                  lambda x: x, equation_order=0)
        signals.add(hysteresis_input)
    #
    return signals

def generate_basis_functions(basis_functions: list[BasisFunction], signals: SignalVector):
    """Generate basis functions using `Signal` class."""
    # Second, we generate basis-function trajectories
    basis_function_signals = []
    for basis_function in basis_functions:
        basic_signal = signals.find(basis_function.variable)
        #
        if basic_signal is None:
            raise ValueError('Symbol not recognized.')
        #
        basis_function_signal = perform_signal_operation(basic_signal, basis_function.operation,
                                                            basis_function.arguments)
        basis_function_signals.append( basis_function_signal )
    #
    return SignalVector(basis_function_signals)

def generate_io_trajectories(io_signal_vector: list[Signal], time_delays, model_order):
    """Generate a dictionary of input-output signal trajectories
    including their delayed versions."""
    io_trajectories_dict = {}
    for n_s, signal in enumerate(io_signal_vector):
        for delay in time_delays[n_s]:
            key = signal.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
            indices = (model_order - delay, -delay if delay > 0 else None)
            io_trajectories_dict[key] = signal.trajectory[indices[0]:indices[-1]]
    #
    return io_trajectories_dict

def generate_basis_trajectories(basis_functions: list[Signal], time_delays, model_order):
    """Generate a dictionary of basis-function trajectories
    including their delayed versions."""
    p_trajectories_dict = {}
    possible_time_delays = sorted(set([d for delays in time_delays for d in delays]))
    for signal in basis_functions:
        for delay in possible_time_delays:
            key = signal.symbol + ('(k-' + str(delay) + ')' if delay > 0 else '(k)')
            indices = (model_order - delay, -delay if delay > 0 else None)
            p_trajectories_dict[key] = signal.trajectory[indices[0]:indices[-1]]
    #
    return p_trajectories_dict

def generate_signal_trajectories(signals: Tuple[list[Signal], ...], model_order: int,
                                 no_of_initial_values: int):
    """Generate all relevant signal trajectories including input-output,
    basis-function as well as hysteresis basis-function trajectories."""
    #
    io_signals, basis_function_signals, hysteresis_basis_function_signals = signals
    #
    time_delays = [np.arange(1, model_order + 1, 1).tolist(), # voltage
                    np.arange(0, model_order + 1, 1).tolist(), # current
                    [0]] # hysteresis input
    #
    io_trajectories_dict = generate_io_trajectories(io_signals,
                                                    time_delays, no_of_initial_values)
    p_trajectories_dict = generate_basis_trajectories(basis_function_signals,
                                                        time_delays, no_of_initial_values)
    h_trajectories_dict = generate_basis_trajectories(hysteresis_basis_function_signals,
                                                        time_delays, no_of_initial_values)
    #
    return io_trajectories_dict, p_trajectories_dict, h_trajectories_dict



from cvxopt import matrix, spdiag, mul, div, sqrt, normal, setseed
from cvxopt import blas, lapack, solvers, sparse, spmatrix
import math

try:
    import mosek
    import sys
    __MOSEK = True
except: __MOSEK = False

if __MOSEK:

    def l1regls_mosek(A, b):
        """

        Returns the solution of l1-norm regularized least-squares problem

            minimize    || A*x - b ||_2^2  + e'*u

            subject to  -u <= x <= u

        """

        m, n = A.size

        env  = mosek.Env()
        task = env.Task(0,0)
        task.set_Stream(mosek.streamtype.log, lambda x: sys.stdout.write(x))

        task.appendvars( 2*n)            # number of variables
        task.appendcons( 2*n)            # number of constraints

        # input quadratic objective
        Q = matrix(0.0, (n,n)) 
        blas.syrk(A, Q, alpha = 2.0, trans='T')

        I = []
        for i in range(n):
            I.extend(range(i,n))

        J = []
        for i in range(n):
            J.extend((n-i)*[i])

        task.putqobj(I, J, list(Q[matrix(I) + matrix(J)*n]))
        task.putclist(range(2*n), list(-2*A.T*b) + n*[1.0])  # setup linear objective

        # input constraint matrix row by row
        for i in range(n):
            task.putarow(   i, [i, n+i], [1.0, -1.0])
            task.putarow( n+i, [i, n+i], [1.0,  1.0])

        # setup bounds on constraints
        task.putboundslice(mosek.accmode.con,
                           0, n, n*[mosek.boundkey.up], n*[0.0], n*[0.0])
        task.putboundslice(mosek.accmode.con,
                           n, 2*n, n*[mosek.boundkey.lo], n*[0.0], n*[0.0])

        # setup variable bounds
        task.putboundslice(mosek.accmode.var,
                           0, 2*n, 2*n*[mosek.boundkey.fr], 2*n*[0.0], 2*n*[0.0])

        # optimize the task
        task.putobjsense(mosek.objsense.minimize)
        task.optimize()
        task.solutionsummary(mosek.streamtype.log)
        x = n*[0.0]
        task.getsolutionslice(mosek.soltype.itr, mosek.solitem.xx, 0, n, x)

        return matrix(x)

    def l1regls_mosek2(A, b):
        """

        Returns the solution of l1-norm regularized least-squares problem

            minimize     w'*w + e'*u

            subject to  -u <= x <= u

                         A*x - w = b

        """

        m, n = A.size

        env  = mosek.Env()
        task = env.Task(0,0)
        task.set_Stream(mosek.streamtype.log, lambda x: sys.stdout.write(x))

        task.appendvars(2*n + m)     # number of variables
        task.appendcons(2*n + m)     # number of constraints

        # input quadratic objective
        task.putqobj(range(2*n,2*n+m), range(2*n,2*n+m), m*[2.0])

        task.putclist(range(2*n+m), n*[0.0] + n*[1.0] + m*[0.0])  # setup linear objective

        # input constraint matrix row by row
        for i in range(n):
            task.putarow(   i, [i, n+i], [1.0, -1.0])
            task.putarow( n+i, [i, n+i], [1.0,  1.0])

        for i in range(m):
            task.putarow( 2*n+i, range(n) + [2*n+i], list(A[i,:]) + [-1.0])

        # setup bounds on constraints
        task.putboundslice(mosek.accmode.con,
                           0, n, n*[mosek.boundkey.up], n*[0.0], n*[0.0])
        task.putboundslice(mosek.accmode.con,
                           n, 2*n, n*[mosek.boundkey.lo], n*[0.0], n*[0.0])
        task.putboundslice(mosek.accmode.con,
                           2*n, 2*n+m, m*[mosek.boundkey.fx], list(b), list(b))

        # setup variable bounds
        task.putboundslice(mosek.accmode.var, 0, 2*n+m, (2*n+m)*[mosek.boundkey.fr], 
                           (2*n+m)*[0.0], (2*n+m)*[0.0])

        # optimize the task
        task.putobjsense(mosek.objsense.minimize)
        task.optimize()
        task.solutionsummary(mosek.streamtype.log)
        x = n*[0.0]
        task.getsolutionslice(mosek.soltype.itr, mosek.solitem.xx, 0, n, x)

        return matrix(x)

def l1regls(A, b):
    """
    
    Returns the solution of l1-norm regularized least-squares problem
  
        minimize || A*x - b ||_2^2  + || x ||_1.

    """

    m, n = A.size
    q = matrix(1.0, (2*n,1))
    q[:n] = -2.0 * A.T * b

    def P(u, v, alpha = 1.0, beta = 0.0 ):
        """
            v := alpha * 2.0 * [ A'*A, 0; 0, 0 ] * u + beta * v 
        """
        v *= beta
        v[:n] += alpha * 2.0 * A.T * (A * u[:n])


    def G(u, v, alpha=1.0, beta=0.0, trans='N'):
        """
            v := alpha*[I, -I; -I, -I] * u + beta * v  (trans = 'N' or 'T')
        """

        v *= beta
        v[:n] += alpha*(u[:n] - u[n:])
        v[n:] += alpha*(-u[:n] - u[n:])

    h = matrix(0.0, (2*n,1))


    # Customized solver for the KKT system 
    #
    #     [  2.0*A'*A  0    I      -I     ] [x[:n] ]     [bx[:n] ]
    #     [  0         0   -I      -I     ] [x[n:] ]  =  [bx[n:] ].
    #     [  I        -I   -D1^-1   0     ] [zl[:n]]     [bzl[:n]]
    #     [ -I        -I    0      -D2^-1 ] [zl[n:]]     [bzl[n:]]
    #
    # where D1 = W['di'][:n]**2, D2 = W['di'][:n]**2.
    #    
    # We first eliminate zl and x[n:]:
    #
    #     ( 2*A'*A + 4*D1*D2*(D1+D2)^-1 ) * x[:n] = 
    #         bx[:n] - (D2-D1)*(D1+D2)^-1 * bx[n:] + 
    #         D1 * ( I + (D2-D1)*(D1+D2)^-1 ) * bzl[:n] - 
    #         D2 * ( I - (D2-D1)*(D1+D2)^-1 ) * bzl[n:]           
    #
    #     x[n:] = (D1+D2)^-1 * ( bx[n:] - D1*bzl[:n]  - D2*bzl[n:] ) 
    #         - (D2-D1)*(D1+D2)^-1 * x[:n]         
    #
    #     zl[:n] = D1 * ( x[:n] - x[n:] - bzl[:n] )
    #     zl[n:] = D2 * (-x[:n] - x[n:] - bzl[n:] ).
    #
    # The first equation has the form
    #
    #     (A'*A + D)*x[:n]  =  rhs
    #
    # and is equivalent to
    #
    #     [ D    A' ] [ x:n] ]  = [ rhs ]
    #     [ A   -I  ] [ v    ]    [ 0   ].
    #
    # It can be solved as 
    #
    #     ( A*D^-1*A' + I ) * v = A * D^-1 * rhs
    #     x[:n] = D^-1 * ( rhs - A'*v ).

    S = matrix(0.0, (m,m))
    Asc = matrix(0.0, (m,n))
    v = matrix(0.0, (m,1))

    def Fkkt(W):

        # Factor 
        #
        #     S = A*D^-1*A' + I 
        #
        # where D = 2*D1*D2*(D1+D2)^-1, D1 = d[:n]**-2, D2 = d[n:]**-2.

        d1, d2 = W['di'][:n]**2, W['di'][n:]**2

        # ds is square root of diagonal of D
        ds = math.sqrt(2.0) * div( mul( W['di'][:n], W['di'][n:]), 
            sqrt(d1+d2) )
        d3 =  div(d2 - d1, d1 + d2)
     
        # Asc = A*diag(d)^-1/2
        Asc = A * spdiag(ds**-1)

        # S = I + A * D^-1 * A'
        blas.syrk(Asc, S)
        S[::m+1] += 1.0 
        lapack.potrf(S)

        def g(x, y, z):

            x[:n] = 0.5 * ( x[:n] - mul(d3, x[n:]) + 
                mul(d1, z[:n] + mul(d3, z[:n])) - mul(d2, z[n:] - 
                mul(d3, z[n:])) )
            x[:n] = div( x[:n], ds) 

            # Solve
            #
            #     S * v = 0.5 * A * D^-1 * ( bx[:n] - 
            #         (D2-D1)*(D1+D2)^-1 * bx[n:] + 
            #         D1 * ( I + (D2-D1)*(D1+D2)^-1 ) * bzl[:n] - 
            #         D2 * ( I - (D2-D1)*(D1+D2)^-1 ) * bzl[n:] )
                
            blas.gemv(Asc, x, v)
            lapack.potrs(S, v)
            
            # x[:n] = D^-1 * ( rhs - A'*v ).
            blas.gemv(Asc, v, x, alpha=-1.0, beta=1.0, trans='T')
            x[:n] = div(x[:n], ds)

            # x[n:] = (D1+D2)^-1 * ( bx[n:] - D1*bzl[:n]  - D2*bzl[n:] ) 
            #         - (D2-D1)*(D1+D2)^-1 * x[:n]         
            x[n:] = div( x[n:] - mul(d1, z[:n]) - mul(d2, z[n:]), d1+d2 )\
                - mul( d3, x[:n] )
                
            # zl[:n] = D1^1/2 * (  x[:n] - x[n:] - bzl[:n] )
            # zl[n:] = D2^1/2 * ( -x[:n] - x[n:] - bzl[n:] ).
            z[:n] = mul( W['di'][:n],  x[:n] - x[n:] - z[:n] ) 
            z[n:] = mul( W['di'][n:], -x[:n] - x[n:] - z[n:] ) 

        return g

    return solvers.coneqp(P, q, G, h, kktsolver = Fkkt)['x'][:n]




"""
A collection of utilities for battery plots.
"""

from typing import Tuple

from matplotlib import pyplot as plt
from matplotlib.axes import Axes

# pylint: disable=too-many-locals,too-many-arguments
def plot_custom(x_y_data_tuples: list[Tuple], figsize: Tuple[int, int]=(10, 3),
                xlabel: str='x', ylabel: str='y', legends: list[str]|None=None,
                linestyles: list[str]|None=None, xaxis_reverse: bool=False,
                ylims: Tuple[int, int]|None=None, xlims: Tuple[int, int]|None=None,
                title: str|None=None, linewidth: float|None=None,
                colors: list[str]|None=None):
    """Plot utility with various options"""
    #
    plt.figure(figsize=figsize)
    axis: Axes = plt.gca()
    #
    for i, (x_data, y_data) in enumerate(x_y_data_tuples):

        x_data_copy = list(x_data)
        y_data_copy = list(y_data)

        if colors is not None and len(colors) > i:
            axis.plot(x_data_copy, y_data_copy,
                      linestyles[i] if linestyles is not None and len(linestyles) > i else '',
                      color=colors[i] if colors is not None and len(colors) > i else None,
                      lw=linewidth if linewidth is not None else 0.7)
        else:
            axis.plot(x_data_copy, y_data_copy,
                      linestyles[i] if linestyles is not None and len(linestyles) > i else '',
                      lw=linewidth if linewidth is not None else 0.7)
    #
    axis.set_xlabel(xlabel)
    axis.set_ylabel(ylabel)
    #
    axis.grid(linestyle='--', lw=0.5)
    #
    if ylims is not None:
        axis.set_ylim(*ylims)
    #
    if xlims is not None:
        axis.set_xlim(*xlims)
    #
    if legends is not None:
        axis.legend(legends)
    #
    if title is not None:
        axis.set_title(title)
    #
    if xaxis_reverse:
        axis.invert_xaxis()
    #
    plt.show()

def plot_time_vs_voltage(time_voltage_data_tuples: list[Tuple], figsize: Tuple[int, int]=(6, 2),
                         legends: list[str]|None=None, linestyles: list[str]|None=None,
                         linewidth: float|None=None, units: Tuple[str, str]=('sec', 'V'),
                         xlims: Tuple[int, int]|None=None, ylims: Tuple[int, int]|None=None,
                         colors: list[str]|None=None):
    """A shortcut for time vs. voltage plots."""
    #
    plot_custom(time_voltage_data_tuples, figsize, xlabel=f'Time [{units[0]}]',
        ylabel=f'Voltage [{units[1]}]', legends=legends, linestyles=linestyles,
        linewidth=linewidth, xlims=xlims, ylims=ylims, colors=colors)

def plot_time_vs_current(time_current_data_tuples: list[Tuple], figsize: Tuple[int, int]=(6, 2),
                         legends: list[str]|None=None, linestyles: list[str]|None=None,
                         linewidth: float|None=None, units: Tuple[str, str]=('sec', 'A'),
                         xlims: Tuple[int, int]|None=None, ylims: Tuple[int, int]|None=None,
                         colors: list[str]|None=None):
    """A shortcut for time vs. current plots."""
    #
    plot_custom(time_current_data_tuples, figsize, xlabel=f'Time [{units[0]}]',
                ylabel=f'Current [{units[1]}]', legends=legends, linestyles=linestyles,
                ylims=ylims, xlims=xlims, linewidth=linewidth, colors=colors)



"""
A collection of utilities related to an identified model.
"""

import numpy as np

from rich.jupyter import print as rich_print
from rich.table import Table

from .dataclasses import Model

def save_to_file(model: Model, path_to_directory: str, description: str):
    """Save a model to file."""
    file_name = f'{description}_n,l={model.model_order},{model.nonlinearity_order}'
    file_path = f'{path_to_directory}/{file_name}'

    np.save(file_path, model) # type: ignore

def load_from_file(file_path: str):
    """Load a model from file."""
    model = np.load(f'{file_path}', allow_pickle=True).item()
    if not isinstance(model, Model):
        raise ValueError('Cannot load model.')
    return model

def print_model_details(model: Model):
    """Print model details."""
    # We print the details as a table.
    table = Table(title=(f"Model order = {model.model_order}; "
                         f"Nonlinearity order = {model.nonlinearity_order}"))
    #
    table.add_column("Model Term")
    table.add_column("Estimated Parameter")
    #
    for estimate, term in zip(model.model_estimate, model.model_terms):
        table.add_row(str(term), str(estimate))
    #
    rich_print(table, markup=False)



"""
A collection of classes (@dataclasses) for various model
features.
"""

from dataclasses import dataclass, field
from typing import Callable, Any

import numpy as np

@dataclass
class VoltageFunction:
    """A callable dataclass for creating voltage function."""
    func: Callable
    soc_values: list[float]

    def __call__(self, soc):
        return self.func(soc)

    def get_soc_values(self) -> list[float]:
        """Get the soc values used to construct the model."""
        return self.soc_values

@dataclass
class Model:
    """Represents a battery model."""
    model_order: int
    nonlinearity_order: int
    model_terms: np.ndarray
    model_estimate: np.ndarray

@dataclass
class BasisFunction:
    """Represents a basis function."""
    variable: str
    operation: Any
    arguments: list[Any]

@dataclass
class Signal:
    """A realization of a signal."""
    symbol: str
    trajectory: list[float]
    equation: Callable[..., float]
    equation_order: float = field(default_factory=lambda: 0)

    def update_trajectory(self, *args):
        """Add a new value to the corresponding
        trajectory."""

        # Add past values to the arguments depending on the
        # order of the difference equation
        past_values = self.trajectory[-self.equation_order:] if self.equation_order > 0 else []
        updated_args = tuple(past_values) + args

        self.trajectory.append( self.equation( *updated_args ) )

@dataclass
class SignalVector:
    """A collection of signals."""
    signals: list[Signal]

    def get_symbols(self):
        """Get symbols."""
        symbols = []
        for signal in self.signals:
            symbols.append( signal.symbol )
        return symbols

    def add(self, signal):
        """Add a signal to the collection."""
        self.signals.append(signal)

    def find(self, symbol):
        """Find a signal using its symbol."""
        result = None
        for signal in self.signals:
            if signal.symbol == symbol:
                result = signal
                break
        if result is None:
            raise ValueError('Signal not found.')
        return result



"""abc"""

import numpy as np

from cvxopt import matrix
from sklearn.linear_model import RidgeCV

from rich.panel import Panel
from rich.jupyter import print as rich_print

from .l1regls import l1regls

def setup_regression(io_trajectories: dict, p_trajectories: dict, h_trajectories: dict,
                     basis_function_keys: list[list[str]],
                     hysteresis_basis_function_keys: list[str]):
    """Setup regression problem by constructing a regression matrix."""
    phi = []
    column_labels = []
    #
    for key, trajectory in io_trajectories.items():
        # extract delay from key
        symbol = key.split("(")[0]
        delay = "(" + key.split("(")[-1]
        #
        phi.append( trajectory )
        column_labels.append(key)

        if symbol == 'h':
            for basis_key in hysteresis_basis_function_keys:
                #
                h_trajectory = h_trajectories[basis_key + delay]
                phi.append( np.prod([trajectory, h_trajectory], axis=0) )

                combined_key = key + "*" + basis_key + delay
                column_labels.append( combined_key )
        else:
            for basis_key in basis_function_keys:
                #
                p_trajectories_list = [ p_trajectories[term + delay] for term in basis_key ]
                phi.append( np.prod([trajectory, *p_trajectories_list], axis=0) )

                combined_key = key + "*" + "*".join([ term + delay for term in basis_key ])
                column_labels.append( combined_key )
    #
    return np.array(phi).T, np.array(column_labels)

def run_optimizer(regression_matrix: np.ndarray, output_vector: np.ndarray, optimizer: str):
    """Run an optimization routine for a regression problem to
    estimate parameters."""
    #
    if optimizer == 'lasso':
        #
        rich_print(Panel(("Performing LASSO using `l1regls.py` from cvxopt.org. "
                          "See the following link for details:"
                          "\nhttps://cvxopt.org/examples/mlbook/l1regls.html")))
        #
        estimate = l1regls(matrix(regression_matrix), matrix(output_vector))
        estimate = np.array(estimate).flatten()
    elif optimizer == 'ridge':
        #
        rich_print(Panel("Performing cross-validated Ridge regression using `sklearn` package."))
        #
        alphas = np.logspace(-3, 1, num=5 )
        model_fit = RidgeCV(fit_intercept=False, cv=10, alphas=alphas).fit(regression_matrix,
                                                                           output_vector)
        estimate = model_fit.coef_
        #
        rich_print(f"Solution found using [bold]alpha = {model_fit.alpha_}")
    else:
        raise ValueError('Unknown optimization routine.')
    #
    return estimate



"""
This file contains the main function(s) to construct
a voltage (e.g., EMF) model.
"""

import numpy as np

from scipy.interpolate import interp1d, PchipInterpolator

from .dataclasses import VoltageFunction

def load_voltage_model(
        soc_values: list[float], voltage_values: list[float],
        extrapolate=False, interpolator='pchip'
    ) -> VoltageFunction:
    """Load the model directly from the (saved) data."""

    # If the SOC direction is descending, we sort it to
    # ascending first.
    if any(np.diff(soc_values) < 0):
        # We reverse the lists
        soc_values = list(soc_values[::-1])
        voltage_values = list(voltage_values[::-1])

    if interpolator == '1d':
        fill_value = 'extrapolate' if extrapolate else np.nan
        bounds_error = not extrapolate
        interpolator = interp1d(
            soc_values, voltage_values, bounds_error=bounds_error,
            fill_value=fill_value, # type: ignore
            assume_sorted=True
        )
    else:
        interpolator = PchipInterpolator(soc_values, voltage_values, extrapolate=extrapolate)

    return VoltageFunction(
        interpolator,
        soc_values
    )



"""Setup for the PyBatteryID package"""

from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

setup(
    name='pybatteryid',
    version='1.0.0',
    description='Data-driven Battery Model Identification in Python',
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/muizabdul29/PyBatteryID",
    packages=find_packages(include=['pybatteryid', 'pybatteryid.*']),
    install_requires=[
        'numpy>=1.20.0',
        'cvxopt',
        'scikit-learn',
        'rich'
    ],
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



"""
Contains class `InputDesign`
"""

import numpy as np

class InputDesign:
    """This class allows generation of current profile employed as
    identification input using the proposed design in [X]. Note
    that the model output is the battery voltage, and the model
    application is assumed to involve constant temperature."""

    _mean_pulse_period: float
    _std_pulse_period: float

    _mean_rest_period: float
    _std_rest_period: float

    _initial_amplitude: float

    _amplitude_plus: float
    _amplitude_minus: float

    _std_random_amplitude: float

    _dc_offset: float

    _no_of_pulses: int

    _upper_limit: float | None
    _lower_limit: float | None

    def __init__(self,
                 mean_pulse_period: float,
                 std_pulse_period: float,
                 mean_rest_period: float,
                 std_rest_period: float,
                 initial_amplitude: float,
                 amplitude_plus: float,
                 amplitude_minus: float,
                 std_random_amplitude: float,
                 dc_offset: float,
                 no_of_pulses: int,
                 upper_safety_limit: float | None = None,
                 lower_safety_limit: float | None = None) -> None:
        """Load required signal parameters.

        Parameters:
        ----------
            mean_pulse_period : float
                `μ_tau` - Mean pulse period.
            std_pulse_period : float
                `σ_tau` - Standard deviation in the pulse period.
            mean_rest_period : float
                `μ_tau0` - Mean rest period.
            std_rest_period : float
                `σ_tau0` - Standard deviation in the rest period.
            initial_amplitude : float
                `alpha0` - Initial amplitude of the pulse train.
            amplitude_plus : float
                `alpha+` - Upper limit on the final amplitude `alpha`.
            amplitude_minus : float
                `alpha-` - Lower limit on the final amplitude `alpha`.
            std_random_amplitude : float
                `σ_xi` - Standard deviation in the randomness added to
                the signal.
            dc_offset : float
                `nu` - DC offset added to the pulse train.
            no_of_pulses : int
                Number of pulses in the signal.
            upper_safety_limit : float | None
                Upper limit on the identification signal for safety reasons.
            lower_safety_limit : float | None
                Lower limit on the identification signal for safety reasons.
        """
        self._mean_pulse_period = mean_pulse_period
        self._std_pulse_period = std_pulse_period

        self._mean_rest_period = mean_rest_period
        self._std_rest_period = std_rest_period

        self._initial_amplitude = initial_amplitude

        self._amplitude_plus = amplitude_plus
        self._amplitude_minus = amplitude_minus

        self._std_random_amplitude = std_random_amplitude

        self._dc_offset = dc_offset

        self._no_of_pulses = no_of_pulses

        self._upper_limit = upper_safety_limit
        self._lower_limit = lower_safety_limit

    def generate_current_profile(self):
        """Generate current profile used as identification input."""
        # First, we generate periods for each pulse.
        pulse_periods = np.round(np.abs(np.random.normal(self._mean_pulse_period,
                                                         self._std_pulse_period,
                                                         self._no_of_pulses)))
        # Time instants (with sampling period = 1 sec)
        time = np.arange(np.sum(pulse_periods))
        # Starting time instants of pulses
        pulse_start_times = np.cumsum(pulse_periods[:-1])
        pulse_start_times = np.insert(pulse_start_times, 0, 0)

        # Generate rectangular shape by setting 1s' between
        # starting times and half of the corresponding period.
        # Later, it is shifted downward by 0.5, and then multiplied
        # with 2 to scale the amplitudes to unity. The minus ensures
        # the discharge pulse phase at the beginning.
        rectangular_shape = (time >= pulse_start_times[:, np.newaxis]) & \
            (time < pulse_start_times[:, np.newaxis] + 0.5 * pulse_periods[:, np.newaxis])
        pulse_train = -2 * (np.sum(rectangular_shape, axis=0 ) - 0.5)

        # Split pulse train into list of phases
        phase_start_times = np.ceil(np.cumsum(np.repeat(pulse_periods, 2) / 2)).astype(int)
        phases = np.split(pulse_train, phase_start_times[:-1])

        # Add rest period following each phase
        rest_periods = np.rint(np.abs(np.random.normal(self._mean_rest_period,
                                                       self._std_rest_period,
                                                       len(phases)))).astype(int)
        # Define the final amplitude of the signal
        # NOTE: This generates random number per pulse, and not per phase,
        # that is, we generate unique random numbers half the number of
        # phases, and then repeat them.
        final_amplitudes = np.repeat(np.random.uniform(self._amplitude_minus,
                                                       self._amplitude_plus,
                                                       int(len(phases) / 2)), 2)
        # We finally construct the identification signal
        identification_signal = []
        for i, phase in enumerate(phases):
            # First, we define a sequence of zeros to add rest period
            # following each pulse.
            rest_signal = np.zeros(rest_periods[i])
            # Now, we add zero-mean random signal for high-frequency
            # excitations.
            zero_mean_random_signal = np.random.normal(0, self._std_random_amplitude, len(phase))

            finalized_phase = final_amplitudes[i] * (
                self._initial_amplitude * phase + self._dc_offset + zero_mean_random_signal
            )

            # Applying safety limits if provided
            if self._upper_limit is not None:
                finalized_phase[finalized_phase > self._upper_limit] = self._upper_limit
            if self._lower_limit is not None:
                finalized_phase[finalized_phase < self._lower_limit] = self._lower_limit

            identification_signal = np.concatenate((identification_signal,
                                                    finalized_phase,
                                                    rest_signal))
        #
        return identification_signal
